#include <stdlib.h>

#include <set>
#include <utility>

#include "bms_parser.h"

#include "bms_util.h"
#include "bms_exception.h"

namespace {
  std::pair<unsigned int, unsigned int> factorization( unsigned int x ) {
    if ( x == 0 || x == 1 ) {
      return std::make_pair( 1, 1 );
    }

    unsigned int quotient = 1;
    unsigned int prime_factors[] = {2, 3, 5, 7};
    for ( unsigned int i = 0; i < sizeof( prime_factors ) / sizeof( unsigned int ); ++i ) {
      while ( x % prime_factors[i] == 0 ) {
        x /= prime_factors[i];
        quotient *= prime_factors[i];
      }
    }
    return std::make_pair( quotient, x );
  }
}


// -- BmsParseFrame ------------------------------------------------------
BmsParseFrame::BmsParseFrame( BmsBms* master, BmsRandomManager& random_manager ) :
parent_( NULL ),
master_( master ),
master_bgm_channel_count_( new std::vector<unsigned int>( BmsConst::BAR_MAX_COUNT ) ),
current_( &random_manager.GetRootData().GetData() ),
current_bgm_channel_count_( new std::vector<unsigned int>( BmsConst::BAR_MAX_COUNT ) ),
current_if_statement_( &random_manager.GetRoot() ),
if_statement_position_( 0 )
{
}


BmsParseFrame::BmsParseFrame( BmsParseFrame*               parent,
                              BmsBms*                      current,
                              BmsRandomIfStatementElement* current_if_statement,
                              bool                         master_take_over ) :
parent_( parent ),
master_( master_take_over ? parent->master_ : NULL ),
master_bgm_channel_count_( master_take_over ? parent->master_bgm_channel_count_ : NULL ),
current_( current ),
current_bgm_channel_count_( new std::vector<unsigned int>( BmsConst::BAR_MAX_COUNT ) ),
current_if_statement_( current_if_statement ),
if_statement_position_( 0 )
{
  *current_bgm_channel_count_ = *parent->current_bgm_channel_count_;
}


BmsParseFrame::~BmsParseFrame()
{
  if ( parent_ == NULL ) {
    delete master_bgm_channel_count_;
  }
  delete current_bgm_channel_count_;
}


// -- BmsParser ----------------------------------------------------------
BmsParser::BmsParser( void ) :
duplicate_header_handler_( NULL ),
duplicate_bar_change_handler_( NULL ),
parse_error_ignore_handler_( NULL ),
generate_random_handler_( NULL ),

is_first_parse_( true ),
data_(),
line_number_( 0 ),
random_manager_( NULL ),
random_value_queue_(),
frame_stack_(),

bar_( 0 ),
channel_( BmsWord::MIN ),
key_( "" ),
value_( "" )
{
}


void
BmsParser::SetDuplicateHeaderHandler( BmsHandler::DuplicateHeader* handler )
{
  duplicate_header_handler_ = handler;
}

void
BmsParser::SetDuplicateBarChangeHandler( BmsHandler::DuplicateBarChange* handler )
{
  duplicate_bar_change_handler_ = handler;
}

void
BmsParser::SetParseErrorIgnoreHandler( BmsHandler::ParseErrorIgnore* handler )
{
  parse_error_ignore_handler_ = handler;
}

void
BmsParser::SetGenerateRandomHandler( BmsHandler::GenerateRandom* handler )
{
  generate_random_handler_ = handler;
}

unsigned int
BmsParser::GetLineNumber( void ) const
{
  return line_number_;
}


void
BmsParser::Parse( const std::string& filename, BmsBms& bms, BmsRandomManager& random_manager )
{
  BmsTextFileReader reader( filename );
  this->Parse( reader, bms, random_manager );
}

void
BmsParser::Parse( BmsTextFileReader& reader, BmsBms& bms, BmsRandomManager& random_manager )
{
  {
    std::string tmp;
    while ( reader.ReadLine( tmp, true ) ) {
      for ( unsigned int i = tmp.length() - 1; i != 0; --i ) {
        if ( tmp[i] == ' ' || tmp[i] == '\t' ) {
          tmp.erase( i );
        }
        else {
          break;
        }
      }
      data_.push_back( tmp );
    }
  }

  is_first_parse_ = true;
  line_number_ = 0;
  random_manager_ = &random_manager;
  random_value_queue_ = std::queue<unsigned int>();
  frame_stack_ = std::stack<BmsParseFrame*>();
  frame_stack_.push( new BmsParseFrame( &bms, random_manager ) );
  std::set<unsigned int> ignore_line_set;

  for ( std::vector<std::string>::const_iterator it = data_.begin(); it != data_.end(); ++it ) {
    ++line_number_;
    try {
      this->ParseDefault( it->c_str() );
    }
    catch ( BmsParseException& e ) {
      BmsHandler::ParseErrorIgnore* handler = parse_error_ignore_handler_;
      if ( handler == NULL ) {
        handler = &BmsHandler::ParseErrorIgnore::DefaultObject;
      }
      if ( NOT( handler->Handle( e, *it ) ) ) {
        throw;
      }
      ignore_line_set.insert( line_number_ );
    }
  }
  if ( frame_stack_.size() > 1 ) {
    throw BmsParseNotFoundEndIfException( line_number_ );
  }

  is_first_parse_ = false;
  line_number_ = 0;
  for ( std::vector<std::string>::const_iterator it = data_.begin(); it != data_.end(); ++it ) {
    ++line_number_;
    if ( ignore_line_set.find( line_number_ ) != ignore_line_set.end() ) {
      continue;
    }
    try {
      this->ParseDefault( it->c_str() );
    }
    catch ( BmsParseException& e ) {
      BmsHandler::ParseErrorIgnore* handler = parse_error_ignore_handler_;
      if ( handler == NULL ) {
        handler = &BmsHandler::ParseErrorIgnore::DefaultObject;
      }
      if ( NOT( handler->Handle( e, *it ) ) ) {
        throw;
      }
    }
  }
  if ( frame_stack_.size() > 1 ) {
    throw BmsParseNotFoundEndIfException( line_number_ );
  }
}

void
BmsParser::ParseDefault( const char* str )
{
  while ( *str == ' ' || *str == '\t' ) {
    ++str;
  }
  switch ( *str ) {
  case '#':
    this->ParseHeaderOrBar( str + 1 );
    break;

  case '\0':
    break;

  default:
    this->ParseComment( str );
    break;
  }
}


void
BmsParser::ParseComment( const char* str )
{
  // ͓ɉȂ
  NOT_USE_VAR( str );
}


void
BmsParser::ParseHeaderOrBar( const char* str )
{
  if ( BmsUtil::IsDigit( *str ) ) {
    this->ParseBar( str );
  }
  else if ( BmsUtil::IsAlphabet( *str ) ) {
    this->ParseHeaderKey( str );
  }
  else {
    throw BmsParseInvalidCharAsHeaderOrBarException( line_number_, std::string( str, 1 ) );
  }
}


void
BmsParser::ParseHeaderKey( const char* str )
{
  key_.clear();
  for ( unsigned int i = 0; ; ++i ) {
    if ( str[i] == ' ' || str[i] == '\t' || str[i] == '\0' ) {
      key_.assign( str, i );
      BmsUtil::StringToUpper( key_ );
      this->ParseHeaderSeparator( str + i );
      break;
    }
    if ( BmsUtil::IsNotHex36( str[i] ) ) {
      throw BmsParseInvalidCharAsHeaderKeyException( line_number_, std::string( str + i, 1 ) );
    }
  }
}


void
BmsParser::ParseHeaderSeparator( const char* str )
{
  while ( *str == ' ' || *str == '\t' ) {
    ++str;
  }
  this->ParseHeaderValue( str );
}


void
BmsParser::ParseHeaderValue( const char* str )
{
  value_.assign( str );

  if ( key_ == "RANDOM" || key_ == "IF" || key_ == "ENDIF" ) {
    this->ParseRandomStatement();
    return;
  }

  if ( is_first_parse_ ) {
    return;
  }

  // WAV o^
  if ( BmsRegistArray::CheckConstruction( "WAV", key_ ) ) {
    this->ParseRegistArrayValue( BmsBms::GetWavArray );
  }
  // BMP o^
  else if ( BmsRegistArray::CheckConstruction( "BMP", key_ ) ) {
    this->ParseRegistArrayValue( BmsBms::GetBmpArray );
  }
  // BPM o^
  else if ( BmsRegistArray::CheckConstruction( "BPM", key_ ) ) {
    this->ParseRegistArrayValue( BmsBms::GetBpmArray );
  }
  // STOP o^
  else if ( BmsRegistArray::CheckConstruction( "STOP", key_ ) ) {
    this->ParseRegistArrayValue( BmsBms::GetStopArray );
  }
  // ʏwb_o^
  else {
    if ( frame_stack_.top()->current_->GetHeaders().IsExist( key_ ) ) {
      try {
        frame_stack_.top()->current_->GetHeaders().Set(
          key_, (duplicate_header_handler_ ? duplicate_header_handler_ : &BmsHandler::DuplicateHeader::DefaultObject)
          ->Handle( key_, frame_stack_.top()->current_->GetHeaders()[key_], value_ ) );
      }
      catch ( BmsHandler::DuplicateHeader::Error ) {
        throw BmsParseDuplicateHeaderException( line_number_, key_ );
      }
    }
    else {
      frame_stack_.top()->current_->GetHeaders().Set( key_, value_ );
    }

    if ( frame_stack_.top()->master_ != NULL ) {
      if ( frame_stack_.top()->master_->GetHeaders().IsExist( key_ ) ) {
        try {
          frame_stack_.top()->master_->GetHeaders().Set(
            key_, (duplicate_header_handler_ ? duplicate_header_handler_ : &BmsHandler::DuplicateHeader::DefaultObject)
            ->Handle( key_, frame_stack_.top()->master_->GetHeaders()[key_], value_ ) );
        }
        catch ( BmsHandler::DuplicateHeader::Error ) {
          throw BmsParseDuplicateHeaderException( line_number_, key_ );
        }
      }
      else {
        frame_stack_.top()->master_->GetHeaders().Set( key_, value_ );
      }
    }
  }
}


void
BmsParser::ParseRegistArrayValue( BmsRegistArray& (BmsBms::*func)( void ) )
{
  BmsWord pos = BmsWord( key_.substr( key_.length() - 2 ) );

  if ( (frame_stack_.top()->current_->*(func))().IsExist( pos ) ) {
    try {
      (frame_stack_.top()->current_->*(func))().Set(
        pos, (duplicate_header_handler_ ? duplicate_header_handler_ : &BmsHandler::DuplicateHeader::DefaultObject)
        ->Handle( key_, (frame_stack_.top()->current_->*(func))()[pos], value_ ) );
    }
    catch ( BmsHandler::DuplicateHeader::Error ) {
      throw BmsParseDuplicateHeaderException( line_number_, key_ );
    }
  }
  else {
    (frame_stack_.top()->current_->*(func))().Set( pos, value_ );
  }

  if ( frame_stack_.top()->master_ != NULL ) {
    if ( (frame_stack_.top()->master_->*(func))().IsExist( pos ) ) {
      try {
        (frame_stack_.top()->master_->*(func))().Set(
          pos, (duplicate_header_handler_ ? duplicate_header_handler_ : &BmsHandler::DuplicateHeader::DefaultObject)
          ->Handle( key_, (frame_stack_.top()->master_->*(func))()[pos], value_ ) );
      }
      catch ( BmsHandler::DuplicateHeader::Error ) {
        throw BmsParseDuplicateHeaderException( line_number_, key_ );
      }
    }
    else {
      (frame_stack_.top()->master_->*(func))().Set( pos, value_ );
    }
  }
}


void
BmsParser::ParseBar( const char* str )
{
  if ( BmsUtil::IsNotDigit( str[0] ) ||
       BmsUtil::IsNotDigit( str[1] ) ||
       BmsUtil::IsNotDigit( str[2] ) ) {
    throw BmsParseInvalidCharAsBarException( line_number_, std::string( str, 3 ) );
  }
  bar_ = atoi( std::string( str, 3 ).c_str() );
  this->ParseChannel( str + 3 );
}

void
BmsParser::ParseChannel( const char* str )
{
  if ( NOT( BmsWord::CheckConstruction( str ) ) ) {
    throw BmsParseInvalidCharAsChannelException( line_number_, std::string( str, 2 ) );
  }
  channel_ = BmsWord( str );
  if ( ( channel_ == BmsWord( "02" ) && is_first_parse_ ) ||
       ( channel_ != BmsWord( "02" ) && NOT( is_first_parse_ ) ) ) {
    this->ParseColon( str + 2 );
  }
}

void
BmsParser::ParseColon( const char* str )
{
  if ( *str != ':' ) {
    throw BmsParseInvalidCharAsColonException( line_number_, std::string( str, 1 ) );
  }
  if ( is_first_parse_ ) {
    this->ParseBarChangeValue( str + 1 );
  }
  else {
    this->ParseObjectArray( str + 1 );
  }
}


void
BmsParser::ParseBarChangeValue( const char* str )
{
  double bar_change_ratio;

  if ( NOT( BmsUtil::StringToFloat( str, &bar_change_ratio ) ) ) {
    throw BmsParseInvalidBarChangeValueException( line_number_, std::string( str ) );
  }
  if ( bar_change_ratio <= 0.0 ) {
    throw BmsParseInvalidBarChangeValueException( line_number_, std::string( str ) );
  }

  {
    double tmp = bar_change_ratio;
    if ( frame_stack_.top()->current_->GetBarManager()[bar_].GetRatio() != 1.0 ) {
      try {
        tmp = (duplicate_bar_change_handler_ ? duplicate_bar_change_handler_ : &BmsHandler::DuplicateBarChange::DefaultObject)
          ->Handle( bar_, frame_stack_.top()->current_->GetBarManager()[bar_].GetRatio(), bar_change_ratio );
      }
      catch ( BmsHandler::DuplicateBarChange::Error ) {
        throw BmsParseDuplicateBarChangeException( line_number_, bar_ );
      }
    }
    unsigned int new_bar_division_count = this->DecideBarDivisionCount(
      frame_stack_.top()->current_->GetBarManager().GetBarDivisionCount(),
      BmsUtil::GetDenominator( tmp ) );
    if ( new_bar_division_count > frame_stack_.top()->current_->GetBarManager().GetBarDivisionCount() ) {
      frame_stack_.top()->current_->MultiplyBarDivisionCount(
        new_bar_division_count / frame_stack_.top()->current_->GetBarManager().GetBarDivisionCount() );
    }
    frame_stack_.top()->current_->GetBarManager()[bar_].SetRatio( tmp );
  }

  if ( frame_stack_.top()->master_ != NULL ) {
    double tmp = bar_change_ratio;
    if ( frame_stack_.top()->master_->GetBarManager()[bar_].GetRatio() != 1.0 ) {
      try {
        tmp = (duplicate_bar_change_handler_ ? duplicate_bar_change_handler_ : &BmsHandler::DuplicateBarChange::DefaultObject)
          ->Handle( bar_, frame_stack_.top()->master_->GetBarManager()[bar_].GetRatio(), bar_change_ratio );
      }
      catch ( BmsHandler::DuplicateBarChange::Error ) {
        throw BmsParseDuplicateBarChangeException( line_number_, bar_ );
      }
    }
    unsigned int new_bar_division_count = this->DecideBarDivisionCount(
      frame_stack_.top()->master_->GetBarManager().GetBarDivisionCount(),
      BmsUtil::GetDenominator( tmp ) );
    if ( new_bar_division_count > frame_stack_.top()->master_->GetBarManager().GetBarDivisionCount() ) {
      frame_stack_.top()->master_->MultiplyBarDivisionCount(
        new_bar_division_count / frame_stack_.top()->master_->GetBarManager().GetBarDivisionCount() );
    }
    frame_stack_.top()->master_->GetBarManager()[bar_].SetRatio( tmp );
  }
}

void
BmsParser::ParseObjectArray( const char* str )
{
  std::vector<BmsWord> word_array;

  if ( *str == '\0' || *str == ' ' || *str == '\t' ) {
    // 擪ŏI̓IuWF
    throw BmsParseNoObjectArrayException( line_number_ );
  }

  for ( unsigned int i = 0; str[i] != '\0'; i += 2 ) {
    if ( BmsWord::CheckConstruction( str + i ) ) {
      word_array.push_back( BmsWord( str + i ) );
    }
    else if ( str[i] == '\0' || str[i] == ' ' || str[i] == '\t' ) {
      // ̋󔒖
      while ( str[i] != '\0' ) {
        if ( str[i] != ' ' && str[i] != '\t' ) {
          throw BmsParseInvalidEndAsObjectArrayException( line_number_ );
        }
        ++i;
      }
      break;
    }
    else {
      throw BmsParseInvalidCharAsObjectArraryException( line_number_, std::string( str + i, str[i + 1] != '\0' ? 2 : 1 ) );
    }
  }

  this->ParseObjectArrayValue( *frame_stack_.top()->current_, *frame_stack_.top()->current_bgm_channel_count_, word_array );
  if ( frame_stack_.top()->master_ != NULL ) {
    this->ParseObjectArrayValue( *frame_stack_.top()->master_, *frame_stack_.top()->master_bgm_channel_count_, word_array );
  }
}


void
BmsParser::ParseObjectArrayValue( BmsBms& bms,
                                  std::vector<unsigned int>& bgm_channel_count,
                                  const std::vector<BmsWord>& word_array )
{
  unsigned int to = bms.GetBarManager().GetChennelPositionByBarNumber( bar_ );
  unsigned int length = bms.GetBarManager()[bar_].GetLength();

  if ( length % word_array.size() != 0 ) {
    unsigned int request_bar_division_count = static_cast<int>(
      static_cast<double>( word_array.size() ) / bms.GetBarManager()[bar_].GetRatio() );
    unsigned int new_bar_division_count = this->DecideBarDivisionCount(
      bms.GetBarManager().GetBarDivisionCount(), request_bar_division_count );
    if ( new_bar_division_count > bms.GetBarManager().GetBarDivisionCount() ) {
      bms.MultiplyBarDivisionCount( new_bar_division_count / bms.GetBarManager().GetBarDivisionCount() );
      to = bms.GetBarManager().GetChennelPositionByBarNumber( bar_ );
      length = bms.GetBarManager()[bar_].GetLength();
    }
  }

  unsigned int channel_dup_count;
  if ( channel_ == BmsWord( "01" ) ) {
    channel_dup_count = bgm_channel_count[bar_];
    ++bgm_channel_count[bar_];
  }
  else {
    channel_dup_count = 0;
  }

  int k = 0;
  for ( double i = static_cast<double>( to );
        static_cast<unsigned int>( i ) < to + length;
        i += static_cast<double>( length ) / static_cast<double>( word_array.size() ) ) {
    bms.GetChannelManager()[channel_][channel_dup_count][static_cast<unsigned int>( i )] = word_array[k];
    //     BmsChannelManager& mg = bms.GetChannelManager();
    //     BmsChannel& ch = mg[channel_];
    //     BmsChannelBuffer& chbf = ch[channel_dup_count];
    //     BmsWord wd = word_array[static_cast<unsigned int>( i )];
    //     chbf[static_cast<unsigned int>( i )] = wd;
    ++k;
  }
}


void
BmsParser::ParseRandomStatement( void )
{
  if ( is_first_parse_ ) {
    if ( key_ == "RANDOM" ) {
      int tmp;
      if ( NOT( BmsUtil::StringToInteger( value_, &tmp, 10 ) ) ) {
        throw BmsParseInvalidRandomValueException( line_number_, key_, value_ );
      }
      if ( tmp <= 0 ) {
        throw BmsParseInvalidRandomValueException( line_number_, key_, value_ );
      }
      unsigned int random_max = static_cast<unsigned int>( tmp );
      if ( generate_random_handler_ == NULL ) {
        random_manager_->SetValue( BmsHandler::GenerateRandomDefault::DefaultObject.Generate( random_max ) );
      }
      else {
        random_manager_->SetValue( generate_random_handler_->Generate( random_max ) );
      }
      random_value_queue_.push( random_manager_->GetValue() );

      frame_stack_.top()->current_if_statement_->MakeNewGenerateRandomElement( random_max );
    }
    else if ( key_ == "IF" ) {
      int tmp;
      if ( NOT( BmsUtil::StringToInteger( value_, &tmp, 10 ) ) ) {
        throw BmsParseInvalidRandomValueException( line_number_, key_, value_ );
      }
      if ( tmp <= 0 ) {
        throw BmsParseInvalidRandomValueException( line_number_, key_, value_ );
      }
      unsigned int prerequisite = static_cast<unsigned int>( tmp );

      BmsRandomIfStatementElement& if_statement = frame_stack_.top()->current_if_statement_->MakeNewIfStatementElement( prerequisite );
      frame_stack_.push( new BmsParseFrame( frame_stack_.top(),
                                            &if_statement.MakeNewDataElement().GetData(),
                                            &if_statement,
                                            prerequisite == random_manager_->GetValue() ) );
    }
    else if ( key_ == "ENDIF" ) {
      if ( frame_stack_.size() <= 1 ) {
        throw BmsParseUnexceptedEndIfException( line_number_ );
      }
      BmsParseFrame* tmp = frame_stack_.top();
      frame_stack_.pop();
      delete tmp;

      frame_stack_.top()->current_ = &frame_stack_.top()->current_if_statement_->MakeNewDataElement().GetData();
    }
  }
  else {
    if ( key_ == "RANDOM" ) {
      frame_stack_.top()->if_statement_position_ += 1;
      BmsRandomGenerateRandomElement* child = frame_stack_.top()->GetChildElementAs<BmsRandomGenerateRandomElement>();
      if ( child == NULL ) {
        throw BMS_INTERNAL_EXCEPTION;
      }
      random_manager_->SetValue( random_value_queue_.front() );
      random_value_queue_.pop();
    }
    else if ( key_ == "IF" ) {
      frame_stack_.top()->if_statement_position_ += 1;
      BmsRandomIfStatementElement* child_if_statement = frame_stack_.top()->GetChildElementAs<BmsRandomIfStatementElement>();
      if ( child_if_statement == NULL ) {
        throw BMS_INTERNAL_EXCEPTION;
      }
      BmsRandomDataElement* child_current = dynamic_cast<BmsRandomDataElement*>( child_if_statement->GetChildren()[0] );
      if ( child_current == NULL ) {
        throw BMS_INTERNAL_EXCEPTION;
      }
      frame_stack_.push( new BmsParseFrame( frame_stack_.top(),
                                            &child_current->GetData(),
                                            child_if_statement,
                                            child_if_statement->GetPrequisite() == random_manager_->GetValue() ) );
    }
    else if ( key_ == "ENDIF" ) {
      if ( frame_stack_.size() <= 1 ) {
        throw BMS_INTERNAL_EXCEPTION;
      }
      BmsParseFrame* tmp = frame_stack_.top();
      frame_stack_.pop();
      delete tmp;

      frame_stack_.top()->if_statement_position_ += 1;
      BmsRandomDataElement* next_if_statement = frame_stack_.top()->GetChildElementAs<BmsRandomDataElement>();
      if ( next_if_statement == NULL ) {
        throw BMS_INTERNAL_EXCEPTION;
      }
      frame_stack_.top()->current_ = &next_if_statement->GetData();
    }
  }
}


unsigned int
BmsParser::DecideBarDivisionCount( unsigned int current, unsigned int request )
{
  if ( current == 0 || request == 0 ) {
    throw BMS_INTERNAL_EXCEPTION;
  }

  std::pair<unsigned int, unsigned int> factors = factorization( BmsUtil::LCM( current, request ) );
  while ( factors.first < factors.second * 4 ) {
    factors.first *= 2;
  }
  if ( factors.first > BmsConst::BAR_DIVISION_COUNT_MAX ) {
    for ( unsigned int i = 1; ; ++i ) {
      if ( factorization( i ).second != 1 ) {
        continue;
      }
      if ( current * i > request * 3 ) {
        return current * i;
      }
      if ( current * i > BmsConst::BAR_DIVISION_COUNT_MAX ) {
        break;
      }
    }
    throw BmsParseTooManyObjectException( line_number_ );
  }
  return factors.first;
}
